@@ -43,10 +43,11 @@ static gint seek_to_track = -1, current_track = -1;
 static gboolean seeking = FALSE, internal_update = FALSE;
 static guint64 slen = GST_CLOCK_TIME_NONE;
 static gfloat vol = 1.0;
+static void set_gst_ui_and_play (void);
 
 static GtkTreeIter current_iter;
 
-static GtkWidget *play_button, *next_menuitem, *prev_menuitem, *reread_menuitem, *seek_scale, *volume_button, *statusbar;
+static GtkWidget *play_button, *next_menuitem, *prev_menuitem, *reread_menuitem, *seek_scale, *volume_button, *statusbar,  *track_listview;
 
 /**
  * Select track number.
@@ -56,45 +57,16 @@ select_track (void)
 {
   GstElement *cd;
 
-  if (seek_to_track == -1) {
-    gint tracks =
-        gtk_tree_model_iter_n_children (GTK_TREE_MODEL (track_store), NULL);
-
-    seek_to_track = 0;
-    while (seek_to_track < tracks) {
-      gboolean do_play;
-      GtkTreeIter next_iter;
-
-      if (!gtk_tree_model_iter_nth_child (GTK_TREE_MODEL (track_store),
-                                          &next_iter, NULL, seek_to_track)) {
-        g_warning (G_STRLOC ": cannot get nth child");
-        return FALSE;
-      }
-      gtk_tree_model_get (GTK_TREE_MODEL (track_store), &next_iter,
-          COLUMN_EXTRACT, &do_play, -1);
-      if (do_play)
-        break;
-      seek_to_track++;
-    }
-    if (seek_to_track == tracks) {
-      seek_to_track = -1;
-      return FALSE;
-    }
-  }
   if (!gtk_tree_model_iter_nth_child (GTK_TREE_MODEL (track_store),
                                      &current_iter, NULL, seek_to_track)) {
     g_warning (G_STRLOC ": cannot get nth child");
     return FALSE;
   }
 
-  gst_element_set_state (pipeline, GST_STATE_PAUSED);
   cd = gst_bin_get_by_name_recurse_up (GST_BIN (pipeline), "cd-source");
+  gst_element_set_state (pipeline, GST_STATE_PAUSED);
   gst_element_seek (pipeline, 1.0, gst_format_get_by_nick ("track"), GST_SEEK_FLAG_FLUSH, GST_SEEK_TYPE_SET, seek_to_track, GST_SEEK_TYPE_NONE, -1);
-  /* get_state neccessary because of bug #326311 */
-  gst_element_get_state (pipeline, NULL, NULL, -1);
   current_track = seek_to_track;
-  seek_to_track = -1;
-
   return TRUE;
 }
 
@@ -120,7 +92,7 @@ pause (void)
 }
 
 /**
- * Stop.
+ * Stop and reset UI.
  */
 static void
 stop (void)
@@ -128,11 +100,22 @@ stop (void)
   if (pipeline != NULL)
     gst_element_set_state (pipeline, GST_STATE_NULL);
 
+  /* TODO: this should be centralised into the state change logic really */
   gtk_widget_set_sensitive (next_menuitem, FALSE);
   gtk_widget_set_sensitive (prev_menuitem, FALSE);
   gtk_widget_set_sensitive (reread_menuitem, TRUE);
-  /* Call this to ensure that the UI is reset */
+  gtk_widget_hide (seek_scale);
+  gtk_widget_hide (volume_button);
   sj_main_set_title (NULL);
+  gtk_statusbar_push (GTK_STATUSBAR (statusbar), 0, "");
+  gtk_button_set_label (GTK_BUTTON (play_button), GTK_STOCK_MEDIA_PLAY);
+  slen = GST_CLOCK_TIME_NONE;
+
+  if (current_track != -1 &&
+      gtk_tree_model_iter_nth_child (GTK_TREE_MODEL (track_store), &current_iter, NULL, current_track))
+    gtk_list_store_set (track_store, &current_iter, COLUMN_STATE, STATE_IDLE, -1);
+
+  current_track = -1;
 }
 
 /**
@@ -174,18 +157,8 @@ cb_hop_track (GstBus *bus, GstMessage *message, gpointer user_data)
     stop ();
     seek_to_track = 0;
   } else {
-    char *title;
     seek_to_track = next_track;
-    gtk_list_store_set (track_store, &current_iter,
-        COLUMN_STATE, STATE_IDLE, -1);
-    select_track ();
-    gtk_list_store_set (track_store, &current_iter,
-        COLUMN_STATE, STATE_PLAYING, -1);
-    gtk_tree_model_get (model,
-        &current_iter, COLUMN_TITLE, &title, -1);
-    sj_main_set_title (title);
-    g_free (title);
-    play ();
+    set_gst_ui_and_play ();
   }
 }
 
@@ -277,15 +250,15 @@ cb_state (GstBus *bus, GstMessage *message, gpointer user_data)
   GstState old_state, new_state;
   GstStateChange transition;
 
-  gst_message_parse_state_changed (message, &old_state, &new_state, NULL);
-  transition = GST_STATE_TRANSITION (old_state, new_state);
-
   /* all pipe elements will receive state transition messages,
    * so we filter those out. This won't be neccessary after
    * the playbin migration */
-  if ((GstElement*)GST_MESSAGE_SRC(message) != pipeline)
+  if ((GstElement*)GST_MESSAGE_SRC (message) != pipeline)
     return;
 
+  gst_message_parse_state_changed (message, &old_state, &new_state, NULL);
+  transition = GST_STATE_TRANSITION (old_state, new_state);
+
   if (transition == GST_STATE_CHANGE_READY_TO_PAUSED) {
     char *title;
     gtk_widget_show (seek_scale);
@@ -440,32 +413,19 @@ on_play_activate (GtkWidget *button, gpointer user_data)
 
   if (is_playing ()) {
     pause ();
-  } else if (pipeline && GST_STATE (pipeline) == GST_STATE_PAUSED && 
-                (current_track == seek_to_track || seek_to_track == -1)) {
+ } else if (pipeline && GST_STATE (pipeline) == GST_STATE_PAUSED && 
+             current_track == seek_to_track) {
     play ();
-  } else if (setup (&err)) {
-    char *title;
-    if (current_track != -1) {
-      if (!gtk_tree_model_iter_nth_child (GTK_TREE_MODEL (track_store),
+  } else if (pipeline && GST_STATE (pipeline) == GST_STATE_PAUSED &&
+			 current_track != seek_to_track) {
+    if (!gtk_tree_model_iter_nth_child (GTK_TREE_MODEL (track_store),
                                           &current_iter, NULL, current_track))
         return;
-    }
-    gtk_list_store_set (track_store, &current_iter,
-                        COLUMN_STATE, STATE_IDLE, -1);
-    if (select_track ()) {
-      gtk_list_store_set (track_store, &current_iter,
-          COLUMN_STATE, STATE_PLAYING, -1);
-      gtk_tree_model_get (GTK_TREE_MODEL (track_store),
-          &current_iter, COLUMN_TITLE, &title, -1);
-      sj_main_set_title (title);
-      g_free (title);
-
-      play ();
-    } else {
-      g_warning (G_STRLOC ": failed to select track");
-      stop ();
-    }
-  } else {
+	set_gst_ui_and_play ();
+  } else if (setup (&err)) {
+	current_track = -1;
+	cb_hop_track (NULL,NULL,NULL);
+  }	else {
     GtkWidget *dialog;
 
     dialog = gtk_message_dialog_new (GTK_WINDOW (main_window), 0,
@@ -491,30 +451,9 @@ on_tracklist_row_activate (GtkTreeView * treeview, GtkTreePath * path,
   if (!gtk_tree_model_get_iter (model, &iter, path))
     return;
   gtk_tree_model_get (model, &iter, COLUMN_NUMBER, &track, -1);
-  if (gtk_list_store_iter_is_valid (GTK_LIST_STORE (model), &current_iter))
-      gtk_list_store_set (track_store, &current_iter, COLUMN_STATE, STATE_IDLE, -1);
-
   if (setup (&err)) {
-    char *title;
     seek_to_track = track - 1;
-
-    if (current_track != -1) {
-      if (!gtk_tree_model_iter_nth_child (GTK_TREE_MODEL (track_store),
-                                          &current_iter, NULL, current_track))
-        return;
-    }
-
-    gtk_list_store_set (track_store, &current_iter,
-        COLUMN_STATE, STATE_IDLE, -1);
-    select_track ();
-    gtk_list_store_set (track_store, &current_iter,
-        COLUMN_STATE, STATE_PLAYING, -1);
-    gtk_tree_model_get (GTK_TREE_MODEL (track_store),
-        &current_iter, COLUMN_TITLE, &title, -1);
-    sj_main_set_title (title);
-    g_free (title);
-    current_iter = iter;
-    play ();
+    set_gst_ui_and_play ();
   } else {
     GtkWidget *dialog;
 
@@ -560,18 +499,29 @@ on_previous_track_activate(GtkWidget *button, gpointer data)
     stop ();
     seek_to_track = 0;
   } else {
-    char *title;
     seek_to_track = prev_track;
-    gtk_list_store_set (track_store, &current_iter,
-        COLUMN_STATE, STATE_IDLE, -1);
-    select_track ();
+    set_gst_ui_and_play ();
+  }
+}
+
+static void
+set_gst_ui_and_play (void)
+{
+  char *title;
+
+  gtk_list_store_set (track_store, &current_iter,
+      COLUMN_STATE, STATE_IDLE, -1);
+  if (select_track ()) {
     gtk_list_store_set (track_store, &current_iter,
         COLUMN_STATE, STATE_PLAYING, -1);
-    gtk_tree_model_get (model,
+    gtk_tree_model_get (GTK_TREE_MODEL (track_store),
         &current_iter, COLUMN_TITLE, &title, -1);
     sj_main_set_title (title);
     g_free (title);
     play ();
+  } else {
+    g_warning (G_STRLOC ": failed to select track");
+    stop ();
   }
 }
 
@@ -646,11 +596,7 @@ on_seek_release (GtkWidget * scale, GdkEventButton * event, gpointer user_data)
   cd = gst_bin_get_by_name_recurse_up (GST_BIN (pipeline), "cd-source");
   seeking = FALSE;
 
-  /* set_state/get_state neccessary because of bug #326311 */
-  gst_element_set_state (pipeline, GST_STATE_PAUSED);
   gst_element_seek (pipeline, 1.0, GST_FORMAT_TIME, GST_SEEK_FLAG_FLUSH, GST_SEEK_TYPE_SET, slen * val, GST_SEEK_TYPE_NONE, -1);
-  gst_element_get_state (pipeline, NULL, NULL, -1);
-  gst_element_set_state (pipeline, GST_STATE_PLAYING);
 
   return FALSE;
 }
@@ -686,4 +632,5 @@ sj_play_init (void)
   seek_scale = glade_xml_get_widget (glade, "seek_scale");
   volume_button = glade_xml_get_widget (glade, "volume_button");
   statusbar = glade_xml_get_widget (glade, "status_bar");
+  track_listview = glade_xml_get_widget (glade, "track_listview");
 }
